/*****************************************************************************
 * Copyright (C) 2014-2025
 *   
 * Michiel Broek <mbroek at mbse dot eu>
 *
 * This file is part of the mbsePi-apps
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * mbsePi-apps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with ThermFerm; see the file COPYING.  If not, write to the Free
 * Software Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *****************************************************************************/

extern int		debug;

#include "thermferm.h"
#include "statetbl.h"
#include "rdconfig.h"
#include "delay.h"
#include "xutil.h"
#include "websocket.h"
#include "simulator.h"

int			my_simulator_state = THREAD_OFF;
int			my_simulator_command = THREAD_OFF;


#ifdef USE_SIMULATOR

extern sys_config	Config;
extern const char	DEVPRESENT[4][6];



static int simulate(void);


/*
 * Return json data for one simulator
 */
json_object *simulator_json_object(simulator_list *simulator)
{
    json_object *root = json_object_new_object();
    json_object_object_add(root, "uuid", json_object_new_string(simulator->uuid));
    json_object_object_add(root, "name", json_object_new_string(simulator->name));
    json_object_object_add(root, "simno", json_object_new_int(simulator->simno));
    json_object_object_add(root, "volume_air", json_object_new_int(simulator->volume_air));
    json_object_object_add(root, "volume_beer", json_object_new_int(simulator->volume_beer));
    json_object *room = json_object_new_object();
    json_object_object_add(room, "tempaddress", json_object_new_string(simulator->room_tempaddress));
    json_object_object_add(room, "temperature", json_object_new_double_fmt(simulator->room_temperature, "%.1f"));
    json_object_object_add(room, "humaddress", json_object_new_string(simulator->room_humaddress));
    json_object_object_add(room, "humidity", json_object_new_double_fmt(simulator->room_humidity, "%.1f"));
    json_object_object_add(root, "room", room);
    json_object *air = json_object_new_object();
    json_object_object_add(air, "address", json_object_new_string(simulator->air_address));
    json_object_object_add(air, "temperature", json_object_new_double_fmt(simulator->air_temperature, "%.4f"));
    json_object_object_add(air, "present", json_object_new_string(DEVPRESENT[simulator->air_present]));
    json_object_object_add(root, "air", air);
    json_object *beer = json_object_new_object();
    json_object_object_add(beer, "address", json_object_new_string(simulator->beer_address));
    json_object_object_add(beer, "temperature", json_object_new_double_fmt(simulator->beer_temperature, "%.4f"));
    json_object_object_add(beer, "present", json_object_new_string(DEVPRESENT[simulator->beer_present]));
    json_object_object_add(root, "beer", beer);
    json_object *beer2 = json_object_new_object();
    json_object_object_add(beer2, "address", json_object_new_string(simulator->beer_address2));
    json_object_object_add(beer2, "temperature", json_object_new_double_fmt(simulator->beer_temperature2, "%.4f"));
    json_object_object_add(beer2, "present", json_object_new_string(DEVPRESENT[simulator->beer_present2]));
    json_object_object_add(root, "beer2", beer2);
    json_object *chiller = json_object_new_object();
    json_object_object_add(chiller, "address", json_object_new_string(simulator->chiller_address));
    json_object_object_add(chiller, "temperature", json_object_new_double_fmt(simulator->chiller_temperature, "%.4f"));
    json_object_object_add(chiller, "present", json_object_new_string(DEVPRESENT[simulator->chiller_present]));
    json_object_object_add(root, "chiller", chiller);

    json_object *cooler = json_object_new_object();
    json_object_object_add(cooler, "address", json_object_new_string(simulator->cooler_address));
    json_object_object_add(cooler, "temperature", json_object_new_double_fmt(simulator->cooler_temp, "%.4f"));
    json_object_object_add(cooler, "time", json_object_new_int(simulator->cooler_time));
    json_object_object_add(cooler, "size", json_object_new_double_fmt(simulator->cooler_size, "%.4f"));
    json_object_object_add(cooler, "present", json_object_new_string(DEVPRESENT[simulator->cooler_present]));
    json_object_object_add(cooler, "power", json_object_new_int(simulator->cooler_power));
    json_object_object_add(root, "cooler", cooler);
    json_object *heater = json_object_new_object();
    json_object_object_add(heater, "address", json_object_new_string(simulator->heater_address));
    json_object_object_add(heater, "temperature", json_object_new_double_fmt(simulator->heater_temp, "%.4f"));
    json_object_object_add(heater, "time", json_object_new_int(simulator->heater_time));
    json_object_object_add(heater, "size", json_object_new_double_fmt(simulator->heater_size, "%.4f"));
    json_object_object_add(heater, "present", json_object_new_string(DEVPRESENT[simulator->heater_present]));
    json_object_object_add(heater, "power", json_object_new_int(simulator->heater_power));
    json_object_object_add(root, "heater", heater);

    json_object *fan = json_object_new_object();
    json_object_object_add(fan, "address", json_object_new_string(simulator->fan_address));
    json_object_object_add(fan, "present", json_object_new_string(DEVPRESENT[simulator->fan_present]));
    json_object_object_add(fan, "power", json_object_new_int(simulator->fan_power));
    json_object_object_add(root, "fan", fan);
    json_object *light = json_object_new_object();
    json_object_object_add(light, "address", json_object_new_string(simulator->light_address));
    json_object_object_add(light, "present", json_object_new_string(DEVPRESENT[simulator->light_present]));
    json_object_object_add(light, "power", json_object_new_int(simulator->light_power));
    json_object_object_add(root, "light", light);

    json_object *door = json_object_new_object();
    json_object_object_add(door, "address", json_object_new_string(simulator->door_address));
    json_object_object_add(door, "present", json_object_new_string(DEVPRESENT[simulator->door_present]));
    json_object_object_add(door, "value", json_object_new_int(simulator->door_value));
    json_object_object_add(root, "door", door);
    json_object *psu = json_object_new_object();
    json_object_object_add(psu, "address", json_object_new_string(simulator->psu_address));
    json_object_object_add(psu, "present", json_object_new_string(DEVPRESENT[simulator->psu_present]));
    json_object_object_add(psu, "value", json_object_new_int(simulator->psu_value));
    json_object_object_add(root, "psu", psu);

    json_object_object_add(root, "frigo_isolation", json_object_new_double_fmt(simulator->frigo_isolation, "%.4f"));
    json_object_object_add(root, "timestamp", json_object_new_int64(simulator->timestamp));
    return root;
}



void simulator_ws(void)
{
    simulator_list	*simulator;

    json_object *root = json_object_new_object();
    json_object_object_add(root, "type", json_object_new_string("simulator"));
    json_object *metric = json_object_new_array();

    for (simulator = Config.simulators; simulator; simulator = simulator->next) {
	json_object_array_add(metric, simulator_json_object(simulator));
    }
    json_object_object_add(root, "metric", metric);
    ws_broadcast((char *)json_object_to_json_string_ext(root, 0));
    json_object_put(root);
}



void *my_simulator_loop(void *threadid)
{
    pid_t	pid = gettid();

    my_simulator_command = THREAD_RUN;
    syslog(LOG_NOTICE, "Thread my_simulator_loop started, pid=%d", pid);

    /*
     * Run the state machine
     */
    simulate();

    syslog(LOG_NOTICE, "Thread my_simulator_loop stopped");
    return 0;
}


SM_DECL(simulate, (char *)"simulator")
SM_STATES
    Init,
    Pause,
    Waiting,
    Run,
    Websocket
SM_NAMES
    (char *)"Init",
    (char *)"Pause",
    (char *)"Waiting",
    (char *)"run",
    (char *)"Websocket"
SM_EDECL
    simulator_list	*simulator;
    time_t		now, last = (time_t)0;
    int			seconds = 0;
    double		k_room_air, sqm_room_air, thick_room_air, air_heat_transfer;
    double		air_change, vhc_air = 0.00121;
    double		air_temp, beer_temp, chiller_temp;
    bool		changed = false;

SM_START(Init)

SM_STATE(Init)

    for (simulator = Config.simulators; simulator; simulator = simulator->next) {
	/*
	 * Heater and cooler have the air temperature
	 */
	simulator->s_heat_temp = simulator->s_cool_temp = simulator->room_temperature;
    }
    my_simulator_state = THREAD_RUN;
    SM_PROCEED(Waiting);

SM_STATE(Pause)

    my_simulator_state = THREAD_PAUSE;
    if (my_simulator_command == THREAD_OFF) {
        SM_SUCCESS;
    } else if (my_simulator_command == THREAD_RUN) {
	my_simulator_state = THREAD_RUN;
        SM_PROCEED(Waiting);
    }
    mDelay(50);

SM_STATE(Waiting)

    if (my_simulator_command == THREAD_OFF) {
	SM_SUCCESS;
    } else if (my_simulator_command == THREAD_PAUSE) {
	SM_PROCEED(Pause);
    }

    now = time(NULL);
    if (now != last) {
	last = now;
	seconds++;
	SM_PROCEED(Run);
    }
    mDelay(50L);

SM_STATE(Run)

    changed = false;
    for (simulator = Config.simulators; simulator; simulator = simulator->next) {
	if (my_simulator_command == THREAD_OFF) {
	    SM_SUCCESS;
	} else if (my_simulator_command == THREAD_PAUSE) {
	    SM_PROCEED(Pause);
	}

	/*
	 * Copy to duplicates
	 */
	air_temp = simulator->air_temperature;
	beer_temp = simulator->beer_temperature;
	chiller_temp = simulator->chiller_temperature;

	/*
	 * First, calculate temperature difference between the room and the air in the
	 * fridge. We use the volume air to roughly calculate the total area between
	 * the in and outside. Calculate the effect and shift the air temperature towards
	 * the room temperature.
	 */
	sqm_room_air = (cbrtl(simulator->volume_air) * cbrtl(simulator->volume_air) * 6) / 100; /* square meters all fridge sides */
	thick_room_air = 0.04;	/* 4 cm walls   */
	k_room_air = 0.03;	/* Polystrene   */
	air_heat_transfer=(k_room_air * sqm_room_air * (simulator->room_temperature - air_temp)) / thick_room_air;
	air_change = (air_heat_transfer / (vhc_air * ((simulator->volume_air - simulator->volume_beer) * 1000))) / 60.0;
	air_temp += air_change;

	/*
	 * If heating, calculate temperature of the heating plate. If heating is off but
	 * the plate is warmer then the air, calculate the cooling down temperature.
	 * Finally, calculate the new air and plate temperature.
	 */
	if (simulator->heater_present == DEVPRESENT_YES && simulator->heater_power >= 50) {
	    if (simulator->s_heat_temp < simulator->heater_temp) {
		simulator->s_heat_temp += 0.05;
		if (simulator->s_heat_temp > simulator->air_temperature)
		    air_temp += ((simulator->s_heat_temp - air_temp) / 100.0);
	    }
	} else {
	    /*
	     * Follow the air temperature
	     */
	    simulator->s_heat_temp -= (simulator->s_heat_temp - air_temp) / 25.0;
	}

	/*
	 * If cooling, calculate temperature of the cooling plate. If cooling is off but
	 * the plate is colder then the air, calculate the warming up temperature.
	 * Finsally, calculate the new air and plate temperature.
	 */
	if (simulator->cooler_present == DEVPRESENT_YES && simulator->cooler_power >= 50) {
	    if (simulator->s_cool_temp > simulator->cooler_temp) {
		simulator->s_cool_temp -= 0.05;
		if (simulator->s_cool_temp < air_temp)
		    air_temp -= ((air_temp - simulator->s_cool_temp) / 100.0);
	    }
	} else {
	    simulator->s_cool_temp -= (simulator->s_cool_temp - air_temp) / 25.0;
	}

	/*
	 * Calculate final temperature of the beer and the air.
	 */
	// Cheap trick, just follow slowly the air temp.
	beer_temp += ((air_temp - beer_temp) / 500.0);
	air_temp += ((beer_temp - air_temp) / 2500.0);
	chiller_temp = simulator->cooler_temp;        // Link these

	/*
	 * Finally update simulated sensors with the new values.
	 * The devices_loop will pickup the values and sets the resolution.
	 */
	if (air_temp != simulator->air_temperature) {
//	    syslog(LOG_NOTICE, "SIM %d:     air %f to %f", simulator->simno, simulator->air_temperature, air_temp);
	    simulator->air_temperature = air_temp;
	    changed = true;
	}
	if (beer_temp != simulator->beer_temperature) {
//	    syslog(LOG_NOTICE, "SIM %d:    beer %f to %f", simulator->simno, simulator->beer_temperature, beer_temp);
            simulator->beer_temperature = beer_temp;
            changed = true;
	}
	if (chiller_temp != simulator->chiller_temperature) {
//	    syslog(LOG_NOTICE, "SIM %d: chiller %f to %f", simulator->simno, simulator->chiller_temperature, chiller_temp);
	    simulator->chiller_temperature = chiller_temp;
	    changed = true;
	}
    }
    SM_PROCEED(Websocket);

SM_STATE(Websocket)

    if (my_simulator_command == THREAD_OFF) {
	SM_SUCCESS;
    } else if (my_simulator_command == THREAD_PAUSE) {
        SM_PROCEED(Pause);
    }
    if (changed) {
	simulator_ws();
	changed = false;
    }
    SM_PROCEED(Waiting);

SM_END

    my_simulator_state = THREAD_OFF;

SM_RETURN


#endif
